@@ -1,40 +1,259 @@
 package RT::Extension::LDAPImport;
 
-our $VERSION = '0.31';
+our $VERSION = '0.33';
 
 use warnings;
 use strict;
 use base qw(Class::Accessor);
-__PACKAGE__->mk_accessors(qw(_ldap _group screendebug _dnlist));
+__PACKAGE__->mk_accessors(qw(_ldap _group screendebug _users));
 use Carp;
 use Net::LDAP;
+use Net::LDAP::Util qw(escape_filter_value);
+use Net::LDAP::Control::Paged;
+use Net::LDAP::Constant qw(LDAP_CONTROL_PAGED);
 use Data::Dumper;
 
 =head1 NAME
 
 RT::Extension::LDAPImport - Import Users from an LDAP store
 
-
 =head1 SYNOPSIS
 
-    use RT::Extension::LDAPImport;
+In C<RT_SiteConfig.pm>:
+
+    Set($LDAPHost,'my.ldap.host')
+    Set($LDAPUser,'me');
+    Set($LDAPPassword,'mypass');
+    Set($LDAPFilter, '(&(cn = users))');
+    Set($LDAPMapping, {Name         => 'uid', # required
+                       EmailAddress => 'mail',
+                       RealName     => 'cn',
+                       WorkPhone    => 'telephoneNumber',
+                       Organization => 'departmentName'});
+    
+    # Add to any existing plugins
+    Set(@Plugins, qw(RT::Extension::LDAPImport));
+    
+    # If you want to sync Groups RT <-> LDAP
+    
+    Set($LDAPGroupBase, 'ou=Groups,o=Our Place');
+    Set($LDAPGroupFilter, '(&(cn = Groups))');
+    Set($LDAPGroupMapping, {Name               => 'cn',
+                            Member_Attr        => 'member',
+                            Member_Attr_Value  => 'dn' });
+
+Running the import:
+
+    # Run a test import
+    /opt/rt4/local/plugins/RT-Extension-LDAPImport/bin/rtldapimport \
+    --debug > ldapimport.debug 2>&1
+    
+    # Run for real, possibly put in cron
+    /opt/rt4/local/plugins/RT-Extension-LDAPImport/bin/rtldapimport \
+    --import
+
+=head1 CONFIGURATION
+
+All of the configuration for the importer goes
+your C<RT_SiteConfig.pm> file. Some of these values pass through
+to L<Net::LDAP> so you can check there for valid values and more
+advanced options.
+
+=over
+
+=item C<< Set($LDAPHost,'our.ldap.host'); >>
+
+Hostname or ldap(s):// uri:
+
+=item C<< Set($LDAPUser, 'uid=foo,ou=users,dc=example,dc=com'); >>
+
+Your LDAP username or DN. If unset, we'll attempt an anonymous bind.
+
+=item C<< Set($LDAPPassword, 'ldap pass'); >>
+
+Your LDAP password.
+
+=item C<< Set($LDAPBase, 'ou=People,o=Our Place'); >>
+
+Base object to search from.
+
+=item C<< Set($LDAPFilter, '(&(cn = users))'); >>
+
+The LDAP search filter to apply (in this case, find all the users).
+
+=item C<< Set($LDAPMapping... >>
+
+    Set($LDAPMapping, {Name         => 'uid',
+                       EmailAddress => 'mail',
+                       RealName     => 'cn',
+                       WorkPhone    => 'telephoneNumber',
+                       Organization => 'departmentName'});
+
+This provides the mapping of attributes in RT to attribute in LDAP.
+Only Name is required for RT.
+
+The LDAP attributes can also be an arrayref of LDAP fields
+
+    WorkPhone => [qw/CompanyPhone Extension/]
+
+which will be concatenated together with a space.
+
+The LDAP attribute can also be a subroutine reference
+that returns either an arrayref or a list of attributes.
+
+The keys in the mapping (i.e. the RT fields, the left hand side) may be a user
+custom field name prefixed with C<UserCF.>, for example C<< 'UserCF.Employee
+Number' => 'employeeId' >>.  Note that this only B<adds> values at the moment,
+which on single value CFs will remove any old value first.  Multiple value CFs
+may behave not quite how you expect.  If the attribute no longer exists on a
+user in LDAP, it will be cleared on the RT side as well.
+
+You may also prefix any RT custom field name with C<CF.> inside your mapping to
+add available values to a Select custom field.  This effectively takes user
+attributes in LDAP and adds the values as selectable options in a CF.  It does
+B<not> set a CF value on any RT object (User, Ticket, Queue, etc).  You might
+use this to populate a ticket Location CF with all the locations of your users
+so that tickets can be associated with the locations in use.
+
+=item C<< Set($LDAPCreatePrivileged, 1); >>
+
+By default users are created as Unprivileged, but you can change this by
+setting C<$LDAPCreatePrivileged> to 1.
+
+=item C<< Set($LDAPGroupName,'My Imported Users'); >>
+
+The RT Group new and updated users belong to. By default, all users
+added or updated by the importer will belong to the 'Imported from LDAP'
+group.
+
+=item C<< Set($LDAPSkipAutogeneratedGroup, 1); >>
+
+Set this to true to prevent users from being automatically
+added to the group configured by C<$LDAPGroupName>.
+
+=item C<< Set($LDAPUpdateUsers, 1); >>
+
+By default, existing users are skipped.  If you
+turn on LDAPUpdateUsers, we will clobber existing
+data with data from LDAP.
+
+=item C<< Set($LDAPUpdateOnly, 1); >>
+
+By default, we create users who don't exist in RT but do
+match your LDAP filter and obey C<$LDAPUpdateUsers> for existing
+users.  This setting updates existing users, overriding
+C<$LDAPUpdateUsers>, but won't create new
+users who are found in LDAP but not in RT.
+
+=item C<< Set($LDAPGroupBase, 'ou=Groups,o=Our Place'); >>
+
+Where to search for groups to import.
+
+=item C<< Set($LDAPGroupFilter, '(&(cn = Groups))'); >>
+
+The search filter to apply.
+
+=item C<< Set($LDAPGroupMapping... >>
+
+    Set($LDAPGroupMapping, {Name               => 'cn',
+                            Member_Attr        => 'member',
+                            Member_Attr_Value  => 'dn' });
+
+A mapping of RT attributes to LDAP attributes to identify group members.
+Name will become the name of the group in RT, in this case pulling
+from the cn attribute on the LDAP group record returned.
+
+C<Member_Attr> is the field in the LDAP group record the importer should
+look at for group members. These values (there may be multiple members)
+will then be compared to the RT user name, which came from the LDAP
+user record.
+
+C<Member_Attr_Value>, which defaults to 'dn', specifies where on the LDAP
+user record the importer should look to compare the member value.
+A match between the member field on the group record and this
+identifier (dn or other LDAP field) on a user record means the
+user will be added to that group in RT.
+
+You can provide a C<Description> key which will be added as the group
+description in RT. The default description is 'Imported from LDAP'.
+
+=item C<< Set($LDAPSizeLimit, 1000); >>
+
+You can set this value if your LDAP server has result size limits.
+
+=back
+
+=head1 Mapping Groups Between RT and LDAP
+
+If you are using the importer, you likely want to manage access via
+LDAP by putting people in groups like 'DBAs' and 'IT Support', but
+also have groups for other non-RT related things. In this case, you
+won't want to create all of your LDAP groups in RT. To limit the groups
+that get mirrored, construct your C<$LDAPGroupFilter> as an OR (|) with
+all of the RT groups you want to mirror from LDAP. For example:
+
+    Set($LDAPGroupBase, 'OU=Groups,OU=Company,DC=COM');
+    Set($LDAPGroupFilter, '(|(CN=DBAs)(CN=IT Support))');
+
+The importer will then import only the groups that match. In this case,
+import means:
+
+=over
+
+=item * Verifying the group is in AD;
+
+=item * Creating the group in RT if it doesn't exist;
+
+=item * Populating the group with the members identified in AD;
+
+=back
+
+The import script will also issue a warning if a user isn't found in RT,
+but this should only happen when testing. When running with --import on,
+users are created before groups are processed, so all users (group
+members) should exist unless there are inconsistencies in your LDAP configuration.
+
+=head1 Running the Import
+
+Executing C<rtldapimport> will run a test that connects to your LDAP server
+and prints out a list of the users found. To see more about these users,
+and to see more general debug information, include the C<--debug> flag.
+
+That debug information is also sent to the RT log with the debug level.
+Errors are logged to the screen and to the RT log.
+
+Executing C<rtldapimport> with the C<--import> flag will cause it to import
+users into your RT database. It is recommended that you make a database
+backup before doing this. If your filters aren't set properly this could
+create a lot of users or groups in your RT instance.
+
+=head1 RT Versions
+
+The importer works with RT 3.8 and newer including RT 4.
+
+It may work with RT 3.6.
+
+=head1 LDAP Filters
+
+The L<ldapsearch|http://www.openldap.org/software/man.cgi?query=ldapsearch&manpath=OpenLDAP+2.0-Release>
+utility in openldap can be very helpful while refining your filters.
 
 =head1 METHODS
 
 =head2 connect_ldap
 
-Relies on the config variables $RT::LDAPHost,
-$RT::LDAPUser and $RT::LDAPPassword being set
+Relies on the config variables C<$RT::LDAPHost>,
+C<$RT::LDAPUser> and C<$RT::LDAPPassword> being set
 in your RT Config files.
 
- Set(LDAPHost,'my.ldap.host')
- Set(LDAPUSER,'me');
- Set(LDAPPassword,'mypass');
+ Set($LDAPHost,'my.ldap.host')
+ Set($LDAPUSER,'me');
+ Set($LDAPPassword,'mypass');
 
 LDAPUser and LDAPPassword can be blank,
 which will cause an anonymous bind.
 
-LDAPHost can be a hostname or an ldap:// ldaps:// uri
+LDAPHost can be a hostname or an ldap:// ldaps:// uri.
 
 =cut
 
@@ -69,7 +288,7 @@ sub connect_ldap {
 
 =head2 run_user_search
 
-Set up the appropriate arguments for a listing of users
+Set up the appropriate arguments for a listing of users.
 
 =cut
 
@@ -84,9 +303,12 @@ sub run_user_search {
 
 =head2 _run_search
 
-Executes a search using the provided base and filter
+Executes a search using the provided base and filter.
 
-Will connect to LDAP server using connect_ldap
+Will connect to LDAP server using C<connect_ldap>.
+
+Returns an array of L<Net::LDAP::Entry> objects, possibly consolidated from
+multiple LDAP pages.
 
 =cut
 
@@ -100,41 +322,79 @@ sub _run_search {
         return;
     }
 
-    $self->_debug("searching with base => '$args{base}' filter => '$args{filter}'");
+    my %search = (
+        base    => $args{base},
+        filter  => $args{filter},
+    );
+    my (@results, $page, $cookie);
 
-    my $result = $ldap->search( base => $args{base},
-                                filter => $args{filter} );
-
-    if ($result->code) {
-        $self->_error("LDAP search failed " . $result->error);
-        return;
+    if ($RT::LDAPSizeLimit) {
+        $page = Net::LDAP::Control::Paged->new( size => $RT::LDAPSizeLimit, critical => 1 );
+        $search{control} = $page;
     }
 
-    $self->_debug("search found ".$result->count." objects");
-    return $result;
+    LOOP: {
+        # Start where we left off
+        $page->cookie($cookie) if $page and $cookie;
 
+        $self->_debug("searching with: " . join(' ', map { "$_ => '$search{$_}'" } sort keys %search));
+
+        my $result = $ldap->search( %search );
+
+        if ($result->code) {
+            $self->_error("LDAP search failed " . $result->error);
+            last;
+        }
+
+        push @results, $result->entries;
+
+        # Short circuit early if we're done
+        last if not $result->count
+             or $result->count < ($RT::LDAPSizeLimit || 0);
+
+        if ($page) {
+            if (my $control = $result->control( LDAP_CONTROL_PAGED )) {
+                $cookie = $control->cookie;
+            } else {
+                $self->_error("LDAP search didn't return a paging control");
+                last;
+            }
+        }
+        redo if $cookie;
+    }
+
+    # Let the server know we're abandoning the search if we errored out
+    if ($cookie) {
+        $self->_debug("Informing the LDAP server we're done with the result set");
+        $page->cookie($cookie);
+        $page->size(0);
+        $ldap->search( %search );
+    }
+
+    $self->_debug("search found ".scalar @results." objects");
+    return @results;
 }
 
 =head2 import_users import => 1|0
 
 Takes the results of the search from run_search
-and maps attributes from LDAP into RT::User attributes
-using $RT::LDAPMapping.
+and maps attributes from LDAP into C<RT::User> attributes
+using C<$RT::LDAPMapping>.
 Creates RT users if they don't already exist.
 
 With no arguments, only prints debugging information.
-Pass import => 1 to actually change data.
+Pass C<--import> to actually change data.
 
-RT::LDAPMapping should be set in your RT_SiteConfig
-file and looks like this.
+C<$RT::LDAPMapping>> should be set in your C<RT_SiteConfig.pm>
+file and look like this.
 
  Set($LDAPMapping, { RTUserField => LDAPField, RTUserField => LDAPField });
 
-RTUserField is the name of a field on an RT::User object
+RTUserField is the name of a field on an C<RT::User> object
 LDAPField can be a simple scalar and that attribute
-will be looked up in LDAP.  
+will be looked up in LDAP.
 
-It can also be an arrayref, in which case each of the 
+It can also be an arrayref, in which case each of the
 elements will be evaluated in turn.  Scalars will be
 looked up in LDAP and concatenated together with a single
 space.
@@ -146,7 +406,7 @@ If it is an arrayref, the values will be concatenated
 together with a single space.
 
 By default users are created as Unprivileged, but you can change this by
-setting $LDAPCreatePrivileged to 1.
+setting C<$LDAPCreatePrivileged> to 1.
 
 =cut
 
@@ -154,8 +414,8 @@ sub import_users {
     my $self = shift;
     my %args = @_;
 
-    my $results = $self->run_user_search;
-    unless ( $results && $results->count ) {
+    my @results = $self->run_user_search;
+    unless ( @results ) {
         $self->_debug("No results found, no import");
         $self->disconnect_ldap;
         return;
@@ -164,18 +424,17 @@ sub import_users {
     my $mapping = $RT::LDAPMapping;
     return unless $self->_check_ldap_mapping( mapping => $mapping );
 
-    $self->_dnlist({});
+    $self->_users({});
 
-    my $done = 0; my $count = $results->count;
-    while (my $entry = $results->shift_entry) {
-        my $user = $self->_build_object( ldap_entry => $entry, skip => qr/(?i)^CF\./, mapping => $mapping );
-        $user->{Name} ||= $user->{EmailAddress};
+    my $done = 0; my $count = scalar @results;
+    while (my $entry = shift @results) {
+        my $user = $self->_build_user_object( ldap_entry => $entry );
         unless ( $user->{Name} ) {
             $self->_warn("No Name or Emailaddress for user, skipping ".Dumper $user);
             next;
         }
         if ( $user->{Name} =~ /^[0-9]+$/) {
-            $self->_warn("Skipping user '$user->{Name}', as it is numeric");
+            $self->_debug("Skipping user '$user->{Name}', as it is numeric");
             next;
         }
         $self->_import_user( user => $user, ldap_entry => $entry, import => $args{import} );
@@ -187,7 +446,7 @@ sub import_users {
 
 =head2 _import_user
 
-The user has run us with --import, so bring data in
+The user has run us with --import, so bring data in.
 
 =cut
 
@@ -196,17 +455,52 @@ sub _import_user {
     my %args = @_;
 
     $self->_debug("Processing user $args{user}{Name}");
-    $self->_dnlist->{lc $args{ldap_entry}->dn} = $args{user}{Name};
+    $self->_cache_user( %args );
 
     $args{user} = $self->create_rt_user( %args );
     return unless $args{user};
 
     $self->add_user_to_group( %args );
     $self->add_custom_field_value( %args );
+    $self->update_object_custom_field_values( %args, object => $args{user} );
 
     return 1;
 }
 
+=head2 _cache_user ldap_entry => Net::LDAP::Entry, [user => { ... }]
+
+Adds the user to a global cache which is used when importing groups later.
+
+Optionally takes a second argument which is a user data object returned by
+_build_user_object.  If not given, _cache_user will call _build_user_object
+itself.
+
+Returns the user Name.
+
+=cut
+
+sub _cache_user {
+    my $self = shift;
+    my %args = (@_);
+    my $user = $args{user} || $self->_build_user_object( ldap_entry => $args{ldap_entry} );
+
+    $self->_users({}) if not defined $self->_users;
+
+    my $group_map       = $RT::LDAPGroupMapping           || {};
+    my $member_attr_val = $group_map->{Member_Attr_Value} || 'dn';
+    my $membership_key  = lc $member_attr_val eq 'dn'
+                            ? $args{ldap_entry}->dn
+                            : $args{ldap_entry}->get_value($member_attr_val);
+
+    # Fallback to the DN if the user record doesn't have a value
+    unless (defined $membership_key) {
+        $membership_key = $args{ldap_entry}->dn;
+        $self->_warn("User attribute '$member_attr_val' has no value for '$membership_key'; falling back to DN");
+    }
+
+    return $self->_users->{lc $membership_key} = $user->{Name};
+}
+
 sub _show_user_info {
     my $self = shift;
     my %args = @_;
@@ -220,7 +514,7 @@ sub _show_user_info {
         my $old_value;
         if ($rt_user) {
             eval { $old_value = $rt_user->$key() };
-            if ($user->{$key} && $old_value eq $user->{$key}) {
+            if ($user->{$key} && defined $old_value && $old_value eq $user->{$key}) {
                 $old_value = 'unchanged';
             }
         }
@@ -232,7 +526,7 @@ sub _show_user_info {
 
 =head2 _check_ldap_mapping
 
-Returns true is there is an LDAPMapping configured,
+Returns true is there is an C<LDAPMapping> configured,
 returns false, logs an error and disconnects from
 ldap if there is no mapping.
 
@@ -253,10 +547,30 @@ sub _check_ldap_mapping {
     return 1;
 }
 
+=head2 _build_user_object
+
+Utility method which wraps C<_build_object> to provide sane
+defaults for building users.  It also tries to ensure a Name
+exists in the returned object.
+
+=cut
+
+sub _build_user_object {
+    my $self = shift;
+    my $user = $self->_build_object(
+        skip    => qr/(?i)^(?:User)?CF\./,
+        mapping => $RT::LDAPMapping,
+        @_
+    );
+    $user->{Name} ||= $user->{EmailAddress};
+    return $user;
+}
+
 =head2 _build_object
 
 Builds up data from LDAP for importing
-Returns a hash of user data ready for RT::User::Create
+Returns a hash of user or group data ready for
+C<RT::User::Create> or C<RT::Group::Create>.
 
 =cut
 
@@ -290,8 +604,8 @@ sub _build_object {
 
 =head3 _parse_ldap_mapping
 
-Internal helper function for import_user
-If we're passed an arrayref, it will recurse 
+Internal helper function for C<import_user>.
+If we're passed an arrayref, it will recurse
 over each of the elements in case one of them is
 another arrayref or subroutine.
 
@@ -324,16 +638,16 @@ sub _parse_ldap_mapping {
 
 =head2 create_rt_user
 
-Takes a hashref of args to pass to RT::User::Create
+Takes a hashref of args to pass to C<RT::User::Create>
 Will try loading the user and will only create a new
-user if it can't find an existing user with the Name
-or EmailAddress arg passed in.
+user if it can't find an existing user with the C<Name>
+or C<EmailAddress> arg passed in.
 
-If the $LDAPUpdateUsers variable is true, data in RT
+If the C<$LDAPUpdateUsers> variable is true, data in RT
 will be clobbered with data in LDAP.  Otherwise we
 will skip to the next user.
 
-If $LDAPUpdateOnly is true, we will not create new users
+If C<$LDAPUpdateOnly> is true, we will not create new users
 but we will update existing ones.
 
 =cut
@@ -404,9 +718,9 @@ sub _load_rt_user {
 
 =head2 add_user_to_group
 
-Adds new users to the group specified in the $LDAPGroupName
-variable (defaults to 'Imported from LDAP')
-You can avoid this if you set $LDAPSkipAutogeneratedGroup
+Adds new users to the group specified in the C<$LDAPGroupName>
+variable (defaults to 'Imported from LDAP').
+You can avoid this if you set C<$LDAPSkipAutogeneratedGroup>.
 
 =cut
 
@@ -442,8 +756,8 @@ sub add_user_to_group {
 
 =head2 setup_group
 
-Pulls the $LDAPGroupName object out of the DB or
-creates it if we ened to do so.
+Pulls the C<$LDAPGroupName> object out of the DB or
+creates it if we need to do so.
 
 =cut
 
@@ -469,7 +783,7 @@ Adds values to a Select (one|many) Custom Field.
 The Custom Field should already exist, otherwise
 this will throw an error and not import any data.
 
-This could probably use some caching
+This could probably use some caching.
 
 =cut
 
@@ -529,18 +843,70 @@ sub add_custom_field_value {
 
 }
 
+=head3 update_object_custom_field_values
+
+Adds CF values to an object (currently only users).  The Custom Field should
+already exist, otherwise this will throw an error and not import any data.
+
+Note that this code only B<adds> values at the moment, which on single value
+CFs will remove any old value first.  Multiple value CFs may behave not quite
+how you expect.
+
+=cut
+
+sub update_object_custom_field_values {
+    my $self = shift;
+    my %args = @_;
+    my $obj  = $args{object};
+
+    foreach my $rtfield ( keys %{$RT::LDAPMapping} ) {
+        # XXX TODO: accept GroupCF when we call this from group_import too
+        next unless $rtfield =~ /^UserCF\.(.+)$/i;
+        my $cf_name = $1;
+        my $ldap_attribute = $RT::LDAPMapping->{$rtfield};
+
+        my @attributes = $self->_parse_ldap_mapping($ldap_attribute);
+        unless (@attributes) {
+            $self->_error("Invalid LDAP mapping for $rtfield ".Dumper($ldap_attribute));
+            next;
+        }
+        my $value = join ' ',
+                    grep { defined and length }
+                     map { scalar $args{ldap_entry}->get_value($_) }
+                         @attributes;
+
+        my $current = $obj->FirstCustomFieldValue($cf_name);
+
+        if (not defined $current and not defined $value) {
+            $self->_debug($obj->Name . ": Skipping '$cf_name'.  No value in RT or LDAP.");
+            next;
+        }
+        elsif (defined $current and defined $value and $current eq $value) {
+            $self->_debug($obj->Name . ": Value '$value' is already set for '$cf_name'");
+            next;
+        }
+
+        $self->_debug($obj->Name . ": Adding object value '$value' for '$cf_name'");
+        next unless $args{import};
+
+        my ($ok, $msg) = $obj->AddCustomFieldValue( Field => $cf_name, Value => $value );
+        $self->_error($obj->Name . ": Couldn't add value '$value' for '$cf_name': $msg")
+            unless $ok;
+    }
+}
+
 =head2 import_groups import => 1|0
 
-Takes the results of the search from run_group_search
-and maps attributes from LDAP into RT::Group attributes
-using $RT::LDAPGroupMapping.
+Takes the results of the search from C<run_group_search>
+and maps attributes from LDAP into C<RT::Group> attributes
+using C<$RT::LDAPGroupMapping>.
 
-Creates groups if they don't exist
+Creates groups if they don't exist.
 
-Removes users from groups if they have been removed from the group on LDAP
+Removes users from groups if they have been removed from the group on LDAP.
 
 With no arguments, only prints debugging information.
-Pass import => 1 to actually change data.
+Pass C<--import> to actually change data.
 
 =cut
 
@@ -548,8 +914,8 @@ sub import_groups {
     my $self = shift;
     my %args = @_;
 
-    my $results = $self->run_group_search;
-    unless ( $results && $results->count ) {
+    my @results = $self->run_group_search;
+    unless ( @results ) {
         $self->_debug("No results found, no group import");
         $self->disconnect_ldap;
         return;
@@ -558,8 +924,8 @@ sub import_groups {
     my $mapping = $RT::LDAPGroupMapping;
     return unless $self->_check_ldap_mapping( mapping => $mapping );
 
-    my $done = 0; my $count = $results->count;
-    while (my $entry = $results->shift_entry) {
+    my $done = 0; my $count = scalar @results;
+    while (my $entry = shift @results) {
         my $group = $self->_build_object( ldap_entry => $entry, skip => qr/(?i)^Member_Attr/, mapping => $mapping );
         $group->{Description} ||= 'Imported from LDAP';
         unless ( $group->{Name} ) {
@@ -567,7 +933,7 @@ sub import_groups {
             next;
         }
         if ( $group->{Name} =~ /^[0-9]+$/) {
-            $self->_warn("Skipping group '$group->{Name}', as it is numeric");
+            $self->_debug("Skipping group '$group->{Name}', as it is numeric");
             next;
         }
         $self->_import_group( %args, group => $group, ldap_entry => $entry );
@@ -579,7 +945,7 @@ sub import_groups {
 
 =head3 run_group_search
 
-Set up the approviate arguments for a listing of users
+Set up the appropriate arguments for a listing of users.
 
 =cut
 
@@ -600,7 +966,7 @@ sub run_group_search {
 
 =head2 _import_group
 
-The user has run us with --import, so bring data in
+The user has run us with C<--import>, so bring data in.
 
 =cut
 
@@ -614,17 +980,18 @@ sub _import_group {
     my ($group_obj, $created) = $self->create_rt_group( %args, group => $group );
     return if $args{import} and not $group_obj;
     $self->add_group_members( %args, name => $group->{Name}, group => $group_obj, ldap_entry => $ldap_entry, new => $created );
+    # XXX TODO: support OCFVs for groups too
     return;
 }
 
 =head2 create_rt_group
 
-Takes a hashref of args to pass to RT::Group::Create
+Takes a hashref of args to pass to C<RT::Group::Create>
 Will try loading the group and will only create a new
-group if it can't find an existing group with the Name
-or EmailAddress arg passed in.
+group if it can't find an existing group with the C<Name>
+or C<EmailAddress> arg passed in.
 
-If $LDAPUpdateOnly is true, we will not create new groups
+If C<$LDAPUpdateOnly> is true, we will not create new groups
 but we will update existing ones.
 
 There is currently no way to prevent Group data from being
@@ -680,7 +1047,7 @@ sub create_rt_group {
 
 =head3 add_group_members
 
-Iterate over the list of DNs in the Member_Attr LDAP entry.
+Iterate over the list of values in the C<Member_Attr> LDAP entry.
 Look up the appropriate username from LDAP.
 Add those users to the group.
 Remove members of the RT Group who are no longer members
@@ -714,23 +1081,28 @@ sub add_group_members {
         $self->_debug("No group in RT, would create with members:");
     }
 
-    my $dnlist = $self->_dnlist;
+    my $users = $self->_users;
     foreach my $member (@$members) {
         my $username;
-        if (exists $dnlist->{lc $member}) {
-            next unless $username = $dnlist->{lc $member};
+        if (exists $users->{lc $member}) {
+            next unless $username = $users->{lc $member};
         } else {
-            my $ldap_users = $self->_run_search(
-                base   => $member,
-                filter => $RT::LDAPFilter,
+            my $attr    = lc($RT::LDAPGroupMapping->{Member_Attr_Value} || 'dn');
+            my $base    = $attr eq 'dn' ? $member : $RT::LDAPBase;
+            my $filter  = $attr eq 'dn'
+                            ? $RT::LDAPFilter
+                            : "(&$RT::LDAPFilter($attr=" . escape_filter_value($member) . "))";
+            my @results = $self->_run_search(
+                base   => $base,
+                filter => $filter,
             );
-            unless ( $ldap_users && $ldap_users->count ) {
-                $dnlist->{lc $member} = undef;
+            unless ( @results ) {
+                $users->{lc $member} = undef;
                 $self->_error("No user found for $member who should be a member of $groupname");
                 next;
             }
-            my $ldap_user = $ldap_users->shift_entry;
-            $dnlist->{lc $member} = $username = $ldap_user->get_value($RT::LDAPMapping->{Name});
+            my $ldap_user = shift @results;
+            $username = $self->_cache_user( ldap_entry => $ldap_user );
         }
         if ( delete $rt_group_members{$username} ) {
             $self->_debug("\t$username\tin RT and LDAP");
@@ -776,7 +1148,7 @@ sub _get_group_members_from_ldap {
 =head2 _show_group
 
 Show debugging information about the group record we're going to import
-when the groups reruns us with --import
+when the groups reruns us with C<--import>.
 
 =cut
 
@@ -810,7 +1182,7 @@ sub _show_group_info {
         my $old_value;
         if ($rt_group) {
             eval { $old_value = $rt_group->$key() };
-            if ($group->{$key} && $old_value eq $group->{$key}) {
+            if ($group->{$key} && defined $old_value && $old_value eq $group->{$key}) {
                 $old_value = 'unchanged';
             }
         }
@@ -822,9 +1194,9 @@ sub _show_group_info {
 
 =head3 disconnect_ldap
 
-Disconnects from the LDAP server
+Disconnects from the LDAP server.
 
-Takes no arguments, returns nothing
+Takes no arguments, returns nothing.
 
 =cut
 
@@ -842,9 +1214,8 @@ sub disconnect_ldap {
 
 =head3 screendebug
 
-We always log to the RT log file with level debug 
-
-This duplicates the messages to the screen
+We always log to the RT log file with level 'debug'. This duplicates
+the messages to the screen.
 
 =cut
 
@@ -891,7 +1262,7 @@ Kevin Falcone  C<< <falcone@bestpractical.com> >>
 
 =head1 LICENCE AND COPYRIGHT
 
-Copyright (c) 2007, Best Practical Solutions, LLC.  All rights reserved.
+Copyright (c) 2007-2012, Best Practical Solutions, LLC.  All rights reserved.
 
 This module is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself. See L<perlartistic>.
